typedef unsigned u32;
typedef unsigned char u8;

#include <algorithm>
#include <omp.h>

/*
    Output resolution: 4260:2400
    Layout: Left 8 channels, Right 8 channels
        Left channels scroll to the left, right channels scroll to the right

    Linux video is 640x408
    Expanded to    852x408
    Scaled to     4260x2400 (original video ends up being 3200x2400)
    Thus, total margin = 4260-3200 = 1060 = 530 on both sides
    We must, thus, render to 530x2400.

    If we want 1080p? 1920x1080
    Then we get, for the margins, 238x1080.

*/

const unsigned extended_xres = 1920, extended_yres = 1080, margin = 238;

class AudioBuffer
{
    static const unsigned rate = 48000, down = 1, length = rate / down;
    static const unsigned per_frame_length = length / 30;//   + length/20;
    static const unsigned sync_room = per_frame_length * 3 / 8;
    static const unsigned render_width = per_frame_length - sync_room;
    unsigned pos_lo, pos_hi;
    float data[length];
    float previous[render_width];
public:
    AudioBuffer(): pos_lo(0), pos_hi(0), data(), previous() { }
    void AddSample(float s)
    {
        data[pos_lo] += s;
        if(++pos_hi >= down) { pos_hi = 0; if(++pos_lo >= length) pos_lo = 0; data[pos_lo] = 0; }
    }
    unsigned FindBeginning()
    {
        //return (pos_lo + length - per_frame_length) % length;
        float    best_diff = 1e30f;
        unsigned chosen_begin = 0;
        omp_lock_t lock; omp_init_lock(&lock);
        //#pragma omp parallel for collapse(2)
        for(unsigned try_begin = 0; try_begin < sync_room; ++try_begin)
            //for(int var=-3; var<=3; ++var)
            {
                //float variation = 1.0 + var*0.01;
                unsigned begin = (pos_lo + length - 1 - per_frame_length + try_begin) % length;
                float prev_s = 0, test_s = 0, avg_diff = 0;
                unsigned index = begin;
                for(unsigned l=0; l<render_width; ++l)
                {
                    //unsigned index = (unsigned)fmod(begin + l*variation, length);
                    float prev_diff = previous[l]/* - prev_s; prev_s = previous[l]*/;
                    float test_diff = data[index]/* - test_s; test_s = data[index]*/;
                    float diff_diff = prev_diff - test_diff;
                    if(diff_diff < 0) diff_diff = -diff_diff;
                    avg_diff += diff_diff * /*diff_diff **/ (render_width-l);
                    //avg_diff += diff_diff /* * diff_diff */ * exp(1+l/float(render_width));//(render_width-l);
                    index = (index+1<length ? index+1 : 0); // go to next sample
                }
                //omp_set_lock(&lock);
                //#pragma omp flush(best_diff,chosen_begin)
                if(avg_diff < best_diff)
                    { best_diff = avg_diff; chosen_begin = begin; }
                //#pragma omp flush(best_diff,chosen_begin)
                //omp_unset_lock(&lock);
            }
        omp_destroy_lock(&lock);
        return chosen_begin;
    }
    void Render(u32* target, unsigned width, unsigned height, bool Reversed)
    {
        const unsigned back = 0x0A0101, line = 0x40200C, spot = 0xE05050;
        //const unsigned opaque = 0xFF000000u;
        const unsigned begin = FindBeginning();

        float lo=1e30f, hi=-1e30f;
        for(unsigned index=begin, p=0; p<render_width; ++p)
        {
            previous[p] = data[index];
            if(data[index] < lo) lo = data[index];
            if(data[index] > hi) hi = data[index];
            index = (index+1<length ? index+1 : 0); // go to next sample
        }
        int mid = height/2;

        float scale    = std::max( lo<0?-lo:lo,  hi<0?-hi:hi );
        float logscale = std::pow(scale, 0.125f);
        if(logscale > 1.f) logscale = 1.f;
        if(scale != 0.f)
            scale = (height/2-1) * logscale / scale;

        auto fade = [=](unsigned x, unsigned color) -> unsigned
        {
            if(!Reversed) x = width-1-x;
            unsigned alpha = 255;
            if(x < width/10)
            {
                alpha = x*256 / (width/10);
            }
            return (alpha << 24) | (color & 0xFFFFFF);
        };
        for(unsigned x=0; x<width; ++x)
        {
            unsigned color = fade(x, back);
            for(unsigned y=0; y<height; ++y)
            {
                u32* p = target + y*extended_xres;
                p[x] = color;
            }
        }
        unsigned index = begin;
        float prev=0;
        for(unsigned p=0; p<render_width; ++p)
        {
            int x = p * width / render_width;
            if(Reversed) x = width-1-x;

            float s = data[index];
            if(p > 0)
            {
                int y1 = mid + scale * prev, y2 = mid + scale * s;
                if(y1 > y2) { int tmp=y1; y1=y2; y2=tmp; }
                if(y1 < 0) y1 = 0;
                if(y2 >= int(height)) y2 = height-1;

                int xx1=x-3; if(xx1<0) xx1=0;
                int xx2=x+3; if(xx2>=int(width)) xx2=width-1;
                for(int xx=xx1; xx<=xx2; ++xx)
                {
                    unsigned color = fade(xx,line);
                    for(int y=y1; y<y2; ++y)
                        target[y * extended_xres + xx] = color;
                }
            }
            prev = s;
            index = (index+1<length ? index+1 : 0); // go to next sample
        }
        index = begin;
        for(unsigned p=0; p<render_width; ++p)
        {
            int x = p * width / render_width;
            if(Reversed) x = width-1-x;

            float s = data[index];
            int y = mid + scale * s;
            int y1 = y-2; if(y1 < 0) y1 = 0;
            int y2 = y+2; if(y2 >= int(height)) y2 = height-1;
            unsigned color = fade(x, spot);
            for(y=y1; y<=y2; ++y)
                target[y * extended_xres + x] = color;
            index = (index+1<length ? index+1 : 0); // go to next sample
        }
    }
} AudioBuffers[16];


static void PutWord(char* target, unsigned dword, int msb_first)
{
    int p;
    for(p=0; p<4; ++p) target[p] = dword >> (msb_first ? 24-p*8 : p*8);
}

static unsigned chksum32(const char* data, unsigned size, unsigned res, int crc)
{
    if(size == 0) return res;
    if(!crc) { unsigned s1 = ((res & 0xFFFF) + (*data & 0xFF)) % 65521; /* adler32 */
    return chksum32(data+1, size-1, s1 + ((((res>>16) + s1) % 65521) << 16), 0); }
    return data ? chksum32(data+1,size-1, ~((~res>>8) ^ chksum32(0,8,(~res ^ *data) & 0xFF,1)), 1)
                : chksum32(0,size-1, (res>>1) ^ ((res&1) ? 0xEDB88320u : 0), 1); /* crc32 */
}

static unsigned adler32(const unsigned char* data, unsigned size, unsigned res)
{
    while(size)
    {
        unsigned s1 = ((res & 0xFFFF) + (*data)) % 65521; /* adler32 */
        ++data;
        --size;
        res = s1 + ((((res>>16) + s1) % 65521) << 16);
    }
    return res;
}


static unsigned Deflate(char* target, const char* source, unsigned srcsize)
{
    /* A bare-bones deflate-compressor (as in gzip) */
    int algo=8, windowbits=8 /* 8 to 15 allowed */, flevel=0 /* 0 to 3 */;
    /* Put RFC1950 header: */
    target[0] = algo + (windowbits-8)*16;
    target[1] = flevel*64 + 31-((256*target[0])%31); /* checksum and compression level */
    /* Compress data using a lossless algorithm (RFC1951): */
    unsigned size = 2; target += 2;

    const char* origsource = source;
    for(unsigned remain=srcsize; remain>0; )
    {
        unsigned eat = remain;
        if(eat > 65535) eat = 65535;;

        target[0] = eat==remain;  /* bfinal bit and btype: 0=uncompressed */

        PutWord(target+1, eat|(~eat<<16), 0);

        size += 5; target += 5;

        memcpy(target, source, eat);

        size += eat; target += eat;
        source += eat;
        remain -= eat;
    }

    /* After compressed data, put the checksum (adler-32): */
    PutWord(target, adler32( (const unsigned char*) origsource, srcsize, 1), 1);
    return size+4;
}

static void EncodeImage(char** target, unsigned w, unsigned h, const u32* frame)
{
    #define AddedChunk(type,n) do { unsigned ncopy=(n); \
        memcpy(*target+4, (type), 4); \
        PutWord(*target, ncopy, 1); /* in the chunk, put crc32 of the type and data (below) */ \
        PutWord(*target + 8 + ncopy, chksum32(*target + 4, 4 + ncopy, 0, 1), 1); \
        *target += 12+ncopy; } while(0)
    std::memcpy(*target, "\x89PNG\15\12\x1A\12", 8); /* Put header - always const */
    *target += 8; // skip header.

      PutWord(*target+8 + 0, w,  1); /* Put image width    */
      PutWord(*target+8 + 4, h,  1); /* and height in IHDR */
      std::memcpy(*target+8 + 8, "\x8\6\0\0\0", 5);
      /* Meaning of above: 8-bit,truecolor+alpha,deflate,std filters,no interlacing */
    AddedChunk("IHDR", 13);

    std::vector<char> framedata( h * (4 * w + 1) );
    for(unsigned p=0,y=0; y<h; ++y)
    {
        framedata[p] = 0;
        memcpy(&framedata[p+1], &frame[y*w], w*4);
        p += 4*w+1;
    }
    AddedChunk("IDAT", Deflate(*target+8, &framedata[0], framedata.size()));

    AddedChunk("IEND", 0);
    #undef AddedChunk
}

const char* audiopat = "frame";
int length_limit = 0x7FFFFFFF, done = 0;

void AudioSetPat(const char* pat, float limit)
{
    audiopat = pat;
    length_limit = limit * 30.f;
}

void Audio_Render()
{
    static u32 frame[extended_xres * extended_yres];
    static char Buffer[sizeof(frame) * 2] = { 0 };

    const unsigned width = margin, height = (extended_yres / 8), inner_height = height*15/16;

    for(auto& u: frame) u = 0u;

    #pragma omp parallel for
    for(unsigned n=0; n<8; ++n)
    {
        AudioBuffers[n].Render(
            frame + (n*height + (height-inner_height)/2) * extended_xres,
            width, inner_height,
            false);
        AudioBuffers[n+8].Render(
            frame + (n*height + (height-inner_height)/2) * extended_xres + extended_xres-width,
            width, inner_height,
            true);
    }

    char* endptr = Buffer;
    EncodeImage(&endptr, extended_xres, extended_yres, frame);

    char cmdbuf[512], fnbuf[64];
    static unsigned frameno=0;
    sprintf(fnbuf, "%.40s%04d.png", audiopat, frameno++);

    FILE* fp = fopen(fnbuf, "wb");
    fwrite(Buffer, endptr-Buffer, 1, fp);
    fclose(fp);

    sprintf(cmdbuf,
        "while [ `sed 's/ .*//;s/\\.//'< /proc/loadavg` -gt 600 ]; do sleep 5;done;"
        "l=`sed 's/ .*//;s/\\.//'< /proc/loadavg`;"
        "if [ $l -lt 450 ]; then "
           "advpng -z2 %s"
         "; else "
           "advpng -z1 %s"
         "; fi &",
        fnbuf, fnbuf);
    system(cmdbuf);

    done += 1;

    if(done >= length_limit)
    {
        fprintf(stderr, "Limit %d met for %s\n", done, audiopat);
        std::exit(0);
    }
}
